import contextvars
from concurrent.futures import ThreadPoolExecutor

default_executor = contextvars.ContextVar("default_executor", default=ThreadPoolExecutor())


class Engine:
    executorClass = ThreadPoolExecutor

    def __init__(self, *args, **kwargs) -> None:
        self.executor = self.executorClass(*args,**kwargs)

    def __enter__(self):
        default_executor.set(self.executor)
        self.executor.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        self.executor.__exit__(*args, **kwargs)
        default_executor.set(None)

    
    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, *args, **kwargs):
        return self.__exit__(*args, **kwargs)
        

class ThreadEngine(Engine):
    executorClass = ThreadPoolExecutor






from herre.graphical import GraphicalBackend, has_webview_error
from herre.console.context import console, get_current_console
from herre.grants.base import BaseGrant
from oauthlib.oauth2 import WebApplicationClient
from requests_oauthlib.oauth2_session import OAuth2Session
from aiohttp import ClientSession
import requests

class AuthorizationCodeGrant(BaseGrant):
    refreshable = True

    def fetchToken(self, **kwargs):

        self.web_app_client = WebApplicationClient(self.config.client_id, scope=self.scope)

        # Create an OAuth2 session for the OSF
        self.session = OAuth2Session(
            self.config.client_id, 
            self.web_app_client,
            scope=self.scope, 
            redirect_uri=self.config.redirect_uri,
            
        )

        with GraphicalBackend():
            assert not has_webview_error, "Please install 'PyQtWebEngine' if you want to use the Implicit Flow"
            from herre.grants.code.widgets.login import LoginDialog
            self.token, accepted = LoginDialog.getToken(backend=self)

        return accepted

   






from herre.exceptions import TokenExpired
from herre.auth import HerreClient
from herre.wards.base import BaseQuery, BaseWard, WardException
import aiohttp
import logging
import re

logger = logging.getLogger(__name__)

gqlparsed_with_variables = re.compile(r"[\s]*(?P<type>subscription|query|mutation)\s*(?P<operation>[a-zA-Z]*)\((?P<arguments>[^\)]*)\)[\s]*{[\s]*(?P<firstchild>[^\(:]*).*")
gqlparser_without_variables = re.compile(r"[\s]*(?P<type>subscription|query|mutation)\s*(?P<operation>[a-zA-Z]*)[\s]*{[\s]*(?P<firstchild>[^\(\{\s:]*).*")

class GQLException(Exception):
    pass



class ParsedQuery(BaseQuery):

    def __init__(self, query: str) -> None:
        self.query = query
        self.variables = None
        self._type = None
        self.m = gqlparsed_with_variables.match(self.query)
        self.has_variables = True
        
        if not self.m:
            self.m = gqlparser_without_variables.match(self.query)
            self.has_variables = False

        if not self.m: raise GQLException(f"Illformed request {self.query}")


    def get_operation_name_for_datakey(self, key):
        #TODO: Make work
        return key

    def combine(self, variables: dict):
        self.variables = variables
        return {"query": self.query, "variables": self.variables}
        
    def parsed(self):
        assert self.variables and self.query, "Please specify query and set variables before parsing"
        return {"query": self.query, "variables": self.variables}

    @property
    def firstchild(self):
        return self.m.group("firstchild")

    @property
    def operation_name(self):
        return self.m.group("operation")

    @property
    def type(self):
        return self.m.group("type")

    def extract(self, result: dict):
        assert self.firstchild in result, f"Cannot Access {self.firstchild}: in {result}"
        assert result[self.firstchild] is not None, f"Empty response {result}"
        return result[self.firstchild]


class GraphQLException(WardException):
    pass

class GraphQLQueryException(GraphQLException):
    pass

class GraphQLProtocolException(WardException):
    pass


class GraphQLWard(BaseWard):

    class Meta:
        abstract = True


    def __init__(self, herre: HerreClient, endpoint: str) -> None:
        super().__init__(herre)
        self.endpoint = endpoint


    async def connect(self):
        if not self.herre.logged_in:
            await self.herre.login()

            
        self.async_session = aiohttp.ClientSession(headers=self.herre.headers)

    async def disconnect(self):
        await self.async_session.close()

    
    async def run(self, gql: ParsedQuery, variables: dict = {}):
        try:
            async with self.async_session.post(self.endpoint, json={"query": gql.query, "variables": variables}) as resp:
                
                if resp.status == 200:
                    result = await resp.json() 
                    logger.debug(f"Received Reply {result}")

                    if "errors" in result:
                        raise GraphQLQueryException(f"Ward {self.endpoint}:" + str(result["errors"]))

                    return result["data"]
                
                if resp.status == 403:
                    logger.error("Auth token is expired trying to refresh")
                    raise TokenExpired("Token Expired Error")

                if resp.status == 400:
                    raise GraphQLProtocolException(await resp.json())

                raise GraphQLProtocolException(f"Unexpected statuscode {resp.status} {resp}")


        except aiohttp.client_exceptions.ClientOSError as e :
            await self.disconnect()
            raise GraphQLProtocolException("Ward is not reachable") from e





from typing import Union
from rich.console import Console as RichConsole
import contextvars


console = contextvars.ContextVar("console", default=None)

def get_current_console() -> RichConsole:
    c = console.get()
    if not c:
        console.set(RichConsole())
    return console.get() 


class Console():

    def __init__(self) -> None:
        self.console = RichConsole()


    def __enter__(self):
        console.set(self)
        return self


    def __exit__(self,*args, **kwargs):
        pass



from herre.config.logging import LoggingConfig
import logging
from rich.logging import RichHandler


def setLogging(config_path=None,**kwargs):

    if config_path:
        config = LoggingConfig.from_file(config_path, overrides=kwargs)
    else:
        config = LoggingConfig(**kwargs)    

    logger = logging.getLogger()
    logger.setLevel(config.level)

    if config.stream:
        stream = RichHandler(markup=True)
        logging.root.setLevel(config.level)
        stream.setLevel(config.level)

        logger.addHandler(stream)

    if config.file:
        file = logging.FileHandler("logs.txt")
        logging.root.setLevel(config.level)
        file.setLevel(config.level)

        logger.addHandler(file)





from .graphical import has_qt_error, has_qasync, QtWidgets, QSelectorEventLoop

def wrap_qt_in_loop(potential_app):
    assert not has_qt_error, f"You do not have the required Librarys installed {has_qt_error}"
    assert not has_qasync, f"You do not have the required Librarys installed {has_qasync}"

    assert isinstance(potential_app, QtWidgets.QApplication), "Passed down app is not an QApplication"
    return QSelectorEventLoop(potential_app)


from herre.auth import get_current_herre
import asyncio
import threading





def loopify(potential_future, as_task=False):
    """Taks an Async Function and according
    to the current loop setting either returns over
    it synchronusly in the herre loop or returns
    a future

    Args:
        potential_future ([type]): [description]

    Returns:
        [type]: [description]
    """

    loop = get_current_herre().loop

    if as_task:
        return loop.create_task(potential_future)

    if loop.is_running():
        if loop._thread_id == threading.current_thread().ident: return potential_future
        return asyncio.run_coroutine_threadsafe(potential_future, loop).result()
    

    return loop.run_until_complete(potential_future)



def next_on_gen(potential_generator, loop):
    """Takes a Async Generator and iterates over it
    threadsafe in the provided loop and returns the result
    synchronously in another generator

    Args:
        potential_generator ([type]): [description]
        loop ([type]): [description]

    Returns:
        [type]: [description]

    Yields:
        [type]: [description]
    """
    ait = potential_generator.__aiter__()

    async def next_on_ait():
        try:
            obj = await ait.__anext__()
            return False, obj
        except StopAsyncIteration:
            return True, None
        
    while True:
        done, obj = asyncio.run_coroutine_threadsafe(next_on_ait(), loop).result()
        if done: 
            break
        yield obj


def loopify_gen(potential_generator):
    """Takes a Async Generator and iterates over it
    threadsafe in the herre loop providing a syncrhonus
    generator or returns an async generator in the current
    loop

    Args:
        potential_generator ([type]): [description]

    Returns:
        [type]: [description]
    """
    loop = get_current_herre().loop
    if loop.is_running():
        if loop._thread_id == threading.current_thread().ident: return potential_generator
        return next_on_gen(potential_generator, loop)



